#include "fd-bg-settings.h"

struct _FdBgSettingsPrivate
{
    GSettings       *gsettings;
    GKeyFile        *global_config_file;
    GFileMonitor    *global_config_monitor;
    gchar           *global_config_filename;

    gchar                       *theme_name;
    gboolean                    use_global_mode;
    guint                       timeout_id;

    gchar                       *image_file;
    GDesktopBackgroundStyle     image_style;
    gint                        image_opacity;
    gint                        delay;
    GDesktopBackgroundShading   color_shading_type;
    gchar                       *primary_color;
    gchar                       *secondary_color;
};

enum
{
    CHANGED,
    LAST_SIGNAL
};
static guint 
background_settings_signals[LAST_SIGNAL] = \
{
    0,
};

G_DEFINE_TYPE_WITH_PRIVATE (FdBgSettings, fd_bg_settings, G_TYPE_OBJECT);

#define GLOBAL_CONFIG           "/etc/desktop/background"
#define GLOBAL_CONFIG_GROUP     "Desktop Background"
#define BACKGROUND_NAME         "name"      //only used in GLOBAL_CONFIG file
#define BACKGROUND_MODE         "mode"      //only used in GLOBAL_CONFIG file

#define PRIMARY_COLOR           "primary-color"
#define SECONDARY_COLOR         "secondary-color"
#define COLOR_SHADING_TYPE      "color-shading-type"
#define IMAGE_STYLE             "picture-options"
#define IMAGE_OPACITY           "picture-opacity"
#define IMAGE_URI               "picture-uri"
#define BACKGROUND_DELAY        "delay"

static void
some_settings_free (FdBgSettings *self)
{
    FdBgSettingsPrivate *priv = self->priv;
    g_free (priv->image_file);
    g_free (priv->primary_color);
    g_free (priv->secondary_color);
}

static void
update_settings_from_gsettings (FdBgSettings *self)
{
    FdBgSettingsPrivate *priv = self->priv;
    some_settings_free (self);
    priv->image_file = g_settings_get_string (priv->gsettings, IMAGE_URI);
    priv->image_style = g_settings_get_enum (priv->gsettings, IMAGE_STYLE);
    priv->image_opacity = g_settings_get_int (priv->gsettings, IMAGE_OPACITY);
    priv->delay = g_settings_get_int (priv->gsettings, BACKGROUND_DELAY);
    priv->color_shading_type = g_settings_get_enum (priv->gsettings, COLOR_SHADING_TYPE);
    priv->primary_color = g_settings_get_string (priv->gsettings, PRIMARY_COLOR);
    priv->secondary_color = g_settings_get_string (priv->gsettings, SECONDARY_COLOR);

    g_signal_emit (self, background_settings_signals[CHANGED], 0);
}

static gboolean
global_config_file_can_write (const gchar *filename)
{
    if (!filename)
        return FALSE;

    if (access (filename, W_OK) != -1)
        return TRUE;
    g_warning ("The config file (%s) has not write permission.", filename);
    return FALSE;
}

static gboolean
save_settings_to_file (FdBgSettings *self)
{
    FdBgSettingsPrivate *priv = self->priv;
    GError *error = NULL;
    if (!priv->global_config_filename)
        return FALSE;
    g_key_file_save_to_file (priv->global_config_file, 
                             priv->global_config_filename, 
                             &error);
    if (error)
    {
        g_warning ("Save background settings failed: %s", error->message);
        g_error_free (error);
        return FALSE;
    }
    return TRUE;
}

static gboolean
check_global_mode_from_string (gchar *mode)
{
    if (g_strcmp0 (mode, "global") == 0)
        return TRUE;
    return FALSE;
}

static GDesktopBackgroundStyle
convert_background_style_from_string (gchar *style)
{
    if (g_strcmp0 (style, "none") == 0)
    {
        return G_DESKTOP_BACKGROUND_STYLE_NONE;
    } 
    else if (g_strcmp0 (style, "wallpaper") == 0)
    {
        return G_DESKTOP_BACKGROUND_STYLE_WALLPAPER;
    }
    else if (g_strcmp0 (style, "centered") == 0)
    {
        return G_DESKTOP_BACKGROUND_STYLE_CENTERED;
    }
    else if (g_strcmp0 (style, "scaled") == 0)
    {
        return G_DESKTOP_BACKGROUND_STYLE_SCALED;
    }
    else if (g_strcmp0 (style, "stretched") == 0)
    {
        return G_DESKTOP_BACKGROUND_STYLE_STRETCHED;
    }
    else if (g_strcmp0 (style, "zoom") == 0)
    {
        return G_DESKTOP_BACKGROUND_STYLE_ZOOM;
    }
    else if (g_strcmp0 (style, "spanned") == 0)
    {
        return G_DESKTOP_BACKGROUND_STYLE_SPANNED;
    }
    return G_DESKTOP_BACKGROUND_STYLE_STRETCHED;
}

static GDesktopBackgroundShading
convert_shading_type_from_string (gchar *shading_type)
{
    if (g_strcmp0 (shading_type, "solid") == 0)
    {
        return G_DESKTOP_BACKGROUND_SHADING_SOLID;
    } 
    else if (g_strcmp0 (shading_type, "vertical") == 0)
    {
        return G_DESKTOP_BACKGROUND_SHADING_VERTICAL;
    }
    else if (g_strcmp0 (shading_type, "horizontal") == 0)
    {
        return G_DESKTOP_BACKGROUND_SHADING_HORIZONTAL;
    }
    return G_DESKTOP_BACKGROUND_SHADING_SOLID;
}

static const gchar *
convert_background_style_to_string (GDesktopBackgroundStyle style)
{
    const gchar *str = NULL;
    switch (style)
    {
        case G_DESKTOP_BACKGROUND_STYLE_NONE:
            str = "none";
            break;
        case G_DESKTOP_BACKGROUND_STYLE_WALLPAPER:
            str = "wallpaper";
            break;
        case G_DESKTOP_BACKGROUND_STYLE_CENTERED:
            str = "centered";
            break;
        case G_DESKTOP_BACKGROUND_STYLE_SCALED:
            str = "scaled";
            break;
        case G_DESKTOP_BACKGROUND_STYLE_ZOOM:
            str = "zoom";
            break;
        case G_DESKTOP_BACKGROUND_STYLE_SPANNED:
            str = "spanned";
            break;
        case G_DESKTOP_BACKGROUND_STYLE_STRETCHED:
        default:
            str = "stretched";
            break;
    }
    return str;
}

static const gchar *
convert_shading_type_to_string (GDesktopBackgroundShading shading_type)
{
    const gchar *str = NULL;
    switch (shading_type)
    {
        case G_DESKTOP_BACKGROUND_SHADING_VERTICAL:
            str = "vertical";
            break;
        case G_DESKTOP_BACKGROUND_SHADING_HORIZONTAL:
            str = "horizontal";
            break;
        case G_DESKTOP_BACKGROUND_SHADING_SOLID:
        default:
            str = "solid";
            break;
    }
    return str;
}

static void
update_image_file_from_global_config (FdBgSettings *self)
{
    FdBgSettingsPrivate *priv = self->priv;
    gchar *str, *uri = NULL;

    str = g_key_file_get_string (priv->global_config_file, 
                                 GLOBAL_CONFIG_GROUP, 
                                 IMAGE_URI, 
                                 NULL);
    if (str)
    {
        if (g_file_test (str, G_FILE_TEST_EXISTS))
        {
            uri = g_strdup (str);
        }
        g_free (str);
    }
    g_debug ("Background image path:%s", uri);
    priv->image_file = uri;
}

static void
update_image_style_from_global_config (FdBgSettings *self)
{
    gchar *style;
    FdBgSettingsPrivate *priv = self->priv;
    style = g_key_file_get_string (priv->global_config_file, 
                                   GLOBAL_CONFIG_GROUP, 
                                   IMAGE_STYLE, 
                                   NULL);
    priv->image_style = convert_background_style_from_string (style);
    g_free (style);
}

static void
update_image_opacity_from_global_config (FdBgSettings *self)
{
    FdBgSettingsPrivate *priv = self->priv;
    gint opacity;
    GError *error = NULL;
    opacity = (int)g_key_file_get_double (priv->global_config_file, 
                                          GLOBAL_CONFIG_GROUP, 
                                          IMAGE_OPACITY, 
                                          &error);
    if (error)
    {
        g_warning ("Parse background image opacity failed:%s", error->message);
        g_error_free (error);
        opacity = 100;  //default value, see gsettings.
    }
    priv->image_opacity = CLAMP(0, opacity, 100);
}

static void
update_background_delay_from_global_config (FdBgSettings *self)
{
    FdBgSettingsPrivate *priv = self->priv;
    gint delay;
    GError *error = NULL;
    delay = (int)g_key_file_get_double (priv->global_config_file, 
                                        GLOBAL_CONFIG_GROUP, 
                                        BACKGROUND_DELAY, 
                                        &error);
    if (error)
    {
        g_warning ("Parse background delay failed:%s", error->message);
        g_error_free (error);
        delay = 600;  //default value, see gsettings.
    }   
    priv->delay = MAX(delay, 0);
}

static void
update_shading_type_from_global_config (FdBgSettings *self)
{
    FdBgSettingsPrivate *priv = self->priv;
    gchar *shading_type;
    shading_type = g_key_file_get_string (priv->global_config_file, 
                                          GLOBAL_CONFIG_GROUP, 
                                          COLOR_SHADING_TYPE, 
                                          NULL);
    priv->color_shading_type = convert_shading_type_from_string (shading_type);
}

static void
update_colors_from_global_config (FdBgSettings *self)
{
    FdBgSettingsPrivate *priv = self->priv;
    priv->primary_color = g_key_file_get_string (priv->global_config_file, 
                                                 GLOBAL_CONFIG_GROUP, 
                                                 PRIMARY_COLOR, 
                                                 NULL);
    priv->secondary_color = g_key_file_get_string (priv->global_config_file, 
                                                   GLOBAL_CONFIG_GROUP, 
                                                   SECONDARY_COLOR, 
                                                   NULL);
    g_debug ("Background primary color: %s, secondary color: %s", priv->primary_color, priv->secondary_color);
}

static void
update_settings_from_global_config (FdBgSettings *self)
{
    FdBgSettingsPrivate *priv = self->priv;
    gchar *name;
    gchar *mode;
    gboolean is_global = FALSE;
    GError *error = NULL;

    if (priv->global_config_filename == NULL)
    {
        update_settings_from_gsettings (self);
        return;
    }

    if (!g_key_file_load_from_file (priv->global_config_file, 
                                    priv->global_config_filename, 
                                    G_KEY_FILE_KEEP_COMMENTS, 
                                    &error))
    {
        g_warning ("Load background config file failed: %s", error->message);
        g_error_free (error);
        g_object_unref (priv->global_config_file);
        priv->global_config_file = NULL;
    }

    if (!priv->global_config_file)
    {
        update_settings_from_gsettings (self);
        return;
    }

    g_debug ("Background settings changed:%s", priv->global_config_filename);
    //update mode;
    mode = g_key_file_get_string (priv->global_config_file, 
                                  GLOBAL_CONFIG_GROUP, 
                                  BACKGROUND_MODE, 
                                  NULL);
    if (mode)
    {
        mode = g_ascii_strdown (mode, -1);
        is_global = check_global_mode_from_string (mode);
    }
    priv->use_global_mode = is_global;

    //update name
    name = g_key_file_get_string (priv->global_config_file, 
                                  GLOBAL_CONFIG_GROUP, 
                                  BACKGROUND_NAME, 
                                  NULL);
    if (name && g_strcmp0 (priv->theme_name, name))
    {
        g_free (priv->theme_name);
        priv->theme_name = g_strdup (name);
    }
    g_free (name);

    //update from gsettings.
    if (priv->use_global_mode == FALSE)
    {
        update_settings_from_gsettings (self);
        return;
    }

    some_settings_free (self);
    update_image_file_from_global_config (self);
    update_image_style_from_global_config (self);
    update_image_opacity_from_global_config (self);
    update_background_delay_from_global_config (self);
    update_shading_type_from_global_config (self);
    update_colors_from_global_config (self);
    g_signal_emit (self, background_settings_signals[CHANGED], 0);
}

static gboolean 
background_settings_update_delay (FdBgSettings *self)
{
    FdBgSettingsPrivate *priv = self->priv;
    priv->timeout_id = 0;
    update_settings_from_global_config (self);
    return FALSE;
}

static void
maybe_background_settings_changed (FdBgSettings *self)
{
    FdBgSettingsPrivate *priv = self->priv;
    
    if (priv->timeout_id > 0)
    {
        g_source_remove (priv->timeout_id);
    }
    /* priv->timeout_id = g_timeout_add (250, 
                                      G_CALLBACK (background_settings_update_delay), 
                                      self); */
    background_settings_update_delay (self);
}

static void
global_config_file_on_changed (GFileMonitor     *monitor,
                                GFile            *file,
                                GFile            *other_file,
                                GFileMonitorEvent event_type,
                                gpointer          user_data)
{
    FdBgSettings *self = FD_BG_SETTINGS (user_data);

    if (event_type == G_FILE_MONITOR_EVENT_CHANGED || 
        event_type == G_FILE_MONITOR_EVENT_DELETED ||
        event_type == G_FILE_MONITOR_EVENT_CREATED)
    {
        maybe_background_settings_changed (self);
    }
}

static void
gsettings_on_changed  (GSettings *settings,
                        char      *key,
                        gpointer   user_data)
{
    FdBgSettings *self = FD_BG_SETTINGS (user_data);
    FdBgSettingsPrivate *priv = self->priv;
    if (priv->use_global_mode == FALSE)
    {
        maybe_background_settings_changed (self);
    }
}

static void
fd_bg_settings_dispose (GObject *gobject)
{
    FdBgSettings *self = FD_BG_SETTINGS (gobject);
    FdBgSettingsPrivate *priv = self->priv;

    if (priv->timeout_id > 0)
    {
        g_source_remove (priv->timeout_id);
        priv->timeout_id = 0;
    }

    if (priv->gsettings)
    {
        g_object_unref (priv->gsettings);
        priv->gsettings = NULL;
    }
    if (priv->global_config_monitor)
    {
        g_object_unref (priv->global_config_monitor);
        priv->global_config_monitor = NULL;
    }
    if (priv->global_config_file)
    {
        g_key_file_unref (priv->global_config_file);
        priv->global_config_file = NULL;
    }

    G_OBJECT_CLASS (fd_bg_settings_parent_class)->dispose (gobject);
}

static void
fd_bg_settings_finalize (GObject *gobject)
{
    FdBgSettings *self = FD_BG_SETTINGS (gobject);
    FdBgSettingsPrivate *priv = self->priv;

    some_settings_free (self);
    g_free (priv->theme_name);
    g_free (priv->global_config_filename);

    G_OBJECT_CLASS (fd_bg_settings_parent_class)->finalize (gobject);
}

static void
fd_bg_settings_class_init (FdBgSettingsClass *klass)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS (klass);

    gobject_class->finalize = fd_bg_settings_finalize;
    gobject_class->dispose = fd_bg_settings_dispose;

    background_settings_signals[CHANGED] = g_signal_new ("changed",
                                                        G_TYPE_FROM_CLASS (klass),
                                                        G_SIGNAL_RUN_LAST,
                                                        NULL,
                                                        NULL,
                                                        NULL,
                                                        g_cclosure_marshal_VOID__VOID,
                                                        G_TYPE_NONE,
                                                        0);
}

static void
fd_bg_settings_init (FdBgSettings *self)
{
    FdBgSettingsPrivate *priv;

    priv = self->priv = fd_bg_settings_get_instance_private (self);
    //Use gsettings as default.
    priv->use_global_mode = FALSE;
    priv->global_config_file = g_key_file_new ();
}

static void
fd_bg_settings_setup (FdBgSettings *self, const gchar *schema_id, const gchar *config_file)
{
    FdBgSettingsPrivate *priv = self->priv;
    GError *error = NULL;
    GFile  *gfile;

    priv->gsettings = g_settings_new (schema_id);
    g_signal_connect (priv->gsettings, 
                      "changed", 
                      G_CALLBACK (gsettings_on_changed), 
                      self);
    priv->global_config_filename = g_strdup (config_file);
    if (config_file)
    {
        gfile = g_file_new_for_path (config_file);
        priv->global_config_monitor = g_file_monitor_file (gfile, 
                                                        G_FILE_MONITOR_NONE, 
                                                        NULL, 
                                                        &error);
        if (error)
        {
            g_warning ("Create config file monitor failed: %s", error->message);
            g_error_free (error);
        }
        g_object_unref (gfile);
        if (priv->global_config_monitor)
        {
            g_signal_connect (priv->global_config_monitor, 
                            "changed", 
                            G_CALLBACK (global_config_file_on_changed), 
                            self);
        }
    }

    update_settings_from_global_config (self);
}

FdBgSettings *
fd_bg_settings_new (const gchar *schema_id, const gchar *config_file)
{
    FdBgSettings *background_settings;
    background_settings = (FdBgSettings *)g_object_new (FD_TYPE_BG_SETTINGS, NULL);
    fd_bg_settings_setup (background_settings, schema_id, config_file);
    return background_settings;
}

gchar *
fd_bg_settings_get_theme_name (FdBgSettings *self)
{
    g_return_val_if_fail (FD_IS_BG_SETTINGS (self), NULL);
    FdBgSettingsPrivate *priv = self->priv;
    return g_strdup (priv->theme_name);
}

gboolean
fd_bg_settings_set_theme_name (FdBgSettings *self, const gchar *name)
{
    g_return_val_if_fail (FD_IS_BG_SETTINGS (self), FALSE);
    FdBgSettingsPrivate *priv = self->priv;
    if (!global_config_file_can_write (priv->global_config_filename))
        return FALSE;

    g_key_file_set_string (priv->global_config_file, 
                           GLOBAL_CONFIG_GROUP, 
                           BACKGROUND_NAME, 
                           name);
    return save_settings_to_file (self);
}

gboolean
fd_bg_settings_is_global_mode (FdBgSettings *self)
{
    FdBgSettingsPrivate *priv = self->priv;
    return priv->use_global_mode;
}

gboolean 
fd_bg_settings_use_global_mode (FdBgSettings *self, gboolean use)
{
    g_return_val_if_fail (FD_IS_BG_SETTINGS (self), FALSE);
    FdBgSettingsPrivate *priv = self->priv;
    const gchar *mode = NULL;
    if (!global_config_file_can_write (priv->global_config_filename))
        return FALSE;

    if (priv->use_global_mode == use)
        return TRUE;

    if (use)
        mode = "global";
    else 
        mode = "default";

    g_key_file_set_string (priv->global_config_file, 
                            GLOBAL_CONFIG_GROUP, 
                            BACKGROUND_MODE, 
                            mode);
    return save_settings_to_file (self);
}

gchar *
fd_bg_settings_get_image_file (FdBgSettings *self)
{
    g_return_val_if_fail (FD_IS_BG_SETTINGS (self), NULL);
    FdBgSettingsPrivate *priv = self->priv;
    return g_strdup (priv->image_file);
}

gboolean
fd_bg_settings_set_image_file (FdBgSettings *self, const gchar *image_file)
{
    g_return_val_if_fail (FD_IS_BG_SETTINGS (self), FALSE);
    FdBgSettingsPrivate *priv = self->priv;
    if (!global_config_file_can_write (priv->global_config_filename))
        return FALSE;

    if (image_file && !g_file_test (image_file, G_FILE_TEST_EXISTS))
    {
        g_warning ("The image file does not exists.");
        return FALSE;
    }
    if (priv->use_global_mode)
    {
        g_key_file_set_string (priv->global_config_file, 
                               GLOBAL_CONFIG_GROUP, 
                               IMAGE_URI, 
                               image_file);
        return save_settings_to_file (self);
    } else {
        g_settings_set_string (priv->gsettings, IMAGE_URI, image_file);
    }
    return TRUE;
}

GDesktopBackgroundStyle
fd_bg_settings_get_image_style (FdBgSettings *self)
{
    g_return_val_if_fail (FD_IS_BG_SETTINGS (self), G_DESKTOP_BACKGROUND_STYLE_STRETCHED);
    FdBgSettingsPrivate *priv = self->priv;
    return priv->image_style;
}

gboolean
fd_bg_settings_set_image_style (FdBgSettings *self, GDesktopBackgroundStyle style)
{
    g_return_val_if_fail (FD_IS_BG_SETTINGS (self), FALSE);
    FdBgSettingsPrivate *priv = self->priv;
    if (!global_config_file_can_write (priv->global_config_filename))
        return FALSE;

    if (priv->use_global_mode)
    {
        const gchar *s = NULL;
        s = convert_background_style_to_string (style);
        g_key_file_set_string (priv->global_config_file, 
                               GLOBAL_CONFIG_GROUP, 
                               IMAGE_STYLE, 
                               s);
        return save_settings_to_file (self);
    } else {
        g_settings_set_enum (priv->gsettings, IMAGE_STYLE, style);
    }
    return TRUE;
}

gint
fd_bg_settings_get_image_opacity (FdBgSettings *self)
{
    g_return_val_if_fail (FD_IS_BG_SETTINGS (self), 100);
    FdBgSettingsPrivate *priv = self->priv;
    return priv->image_style;
}

gboolean
fd_bg_settings_set_image_opacity (FdBgSettings *self, gint opacity)
{
    g_return_val_if_fail (FD_IS_BG_SETTINGS (self), FALSE);
    FdBgSettingsPrivate *priv = self->priv;
    if (priv->use_global_mode)
    {
        g_key_file_set_integer (priv->global_config_file, 
                                GLOBAL_CONFIG_GROUP, 
                                IMAGE_OPACITY, 
                                opacity);
        return save_settings_to_file (self);
    } else {
        g_settings_set_int (priv->gsettings, IMAGE_OPACITY, opacity);
    }
    return TRUE;
}

gint
fd_bg_settings_get_delay (FdBgSettings *self)
{
    g_return_val_if_fail (FD_IS_BG_SETTINGS (self), 600);
    FdBgSettingsPrivate *priv = self->priv;
    return priv->delay;
}

gboolean
fd_bg_settings_set_delay (FdBgSettings *self, gint delay)
{
    g_return_val_if_fail (FD_IS_BG_SETTINGS (self), FALSE);
    FdBgSettingsPrivate *priv = self->priv;
    if (!global_config_file_can_write (priv->global_config_filename))
        return FALSE;

    if (priv->use_global_mode)
    {
        g_key_file_set_integer (priv->global_config_file, 
                                GLOBAL_CONFIG_GROUP, 
                                BACKGROUND_DELAY, 
                                delay);
        return save_settings_to_file (self);
    } else {
        g_settings_set_int (priv->gsettings, BACKGROUND_DELAY, delay);
    }
}

GDesktopBackgroundShading
fd_bg_settings_get_color_shading_type (FdBgSettings *self)
{
    g_return_val_if_fail (FD_IS_BG_SETTINGS (self), G_DESKTOP_BACKGROUND_SHADING_SOLID);
    FdBgSettingsPrivate *priv = self->priv;
    return priv->color_shading_type;
}

gboolean
fd_bg_settings_set_color_shading_type (FdBgSettings *self, GDesktopBackgroundShading shading_type)
{
    g_return_val_if_fail (FD_IS_BG_SETTINGS (self), FALSE);
    FdBgSettingsPrivate *priv = self->priv;
    if (!global_config_file_can_write (priv->global_config_filename))
        return FALSE;

    if (priv->use_global_mode)
    {
        const gchar *s = NULL;
        s = convert_shading_type_to_string (shading_type);
        g_key_file_set_string (priv->global_config_file, 
                               GLOBAL_CONFIG_GROUP, 
                               COLOR_SHADING_TYPE, 
                               s);
        return save_settings_to_file (self);
    } else {
        g_settings_set_enum (priv->gsettings, COLOR_SHADING_TYPE, shading_type);
    }
    return TRUE;
}

gchar *
fd_bg_settings_get_primary_color (FdBgSettings *self)
{
    g_return_val_if_fail (FD_IS_BG_SETTINGS (self), NULL);
    FdBgSettingsPrivate *priv = self->priv;
    return g_strdup (priv->primary_color);
}

gboolean
fd_bg_settings_set_primary_color (FdBgSettings *self, const gchar *color)
{
    g_return_val_if_fail (FD_IS_BG_SETTINGS (self), FALSE);
    FdBgSettingsPrivate *priv = self->priv;
    if (!global_config_file_can_write (priv->global_config_filename))
        return FALSE;
    if (priv->use_global_mode)
    {
        g_key_file_set_string (priv->global_config_file, 
                               GLOBAL_CONFIG_GROUP, 
                               PRIMARY_COLOR, 
                               color);
        return save_settings_to_file (self);
    } else {
        g_settings_set_string (priv->gsettings, PRIMARY_COLOR, color);
    }
    return TRUE;
}

gchar *
fd_bg_settings_get_secondary_color (FdBgSettings *self)
{
    g_return_val_if_fail (FD_IS_BG_SETTINGS (self), NULL);
    FdBgSettingsPrivate *priv = self->priv;
    return g_strdup (priv->secondary_color);
}

gboolean
fd_bg_settings_set_secondary_color (FdBgSettings *self, const gchar *color)
{
    g_return_val_if_fail (FD_IS_BG_SETTINGS (self), FALSE);
    FdBgSettingsPrivate *priv = self->priv;
    if (!global_config_file_can_write (priv->global_config_filename))
        return FALSE;
    if (priv->use_global_mode)
    {
        g_key_file_set_string (priv->global_config_file, 
                               GLOBAL_CONFIG_GROUP, 
                               SECONDARY_COLOR, 
                               color);
        return save_settings_to_file (self);
    } else {
        g_settings_set_string (priv->gsettings, SECONDARY_COLOR, color);
    }
    return TRUE;
}
