#include <stdio.h>
#include "config.h"
#include "fd-idle-action-settings.h"
#include "fd.h"

struct _FdIdleActionSettingsPrivate
{
    GSettings       *power_gsettings;
    GSettings       *session_gsettings;
};

enum
{
    CHANGED,
    LAST_SIGNAL
};

static guint 
settings_signals[LAST_SIGNAL] = {0,};

G_DEFINE_TYPE_WITH_PRIVATE (FdIdleActionSettings, fd_idle_action_settings, G_TYPE_OBJECT);

#define POWER_SCHEMA_ID                     "org.cdos.settings-daemon.plugins.power"
#define SCREENSAVER_SCHEMA_ID               "org.cdos.desktop.session"

#define SCREENSAVER_TIMEOUT                 "idle-screensaver-timeout"
#define SCREENSAVER_LOCK_TIMEOUT            "idle-delay"                //from session
#define SUSPEND_AC_TIMEOUT                  "sleep-inactive-ac-timeout"
#define SUSPEND_BATTERY_TIMEOUT             "sleep-inactive-batter-timeout"
#define DISPLAY_AC_TIMEOUT                  "sleep-display-ac"
#define DISPLAY_BATTERY_TIMEOUT             "sleep-display-battery"


static FdIdleActionType
convert_key_to_action_type (gchar *key)
{
    FdIdleActionType action_type = FD_IDLE_ACTION_NONE;
    if (g_strcmp0 (key, SCREENSAVER_TIMEOUT) == 0)
        action_type = FD_IDLE_ACTION_SCREENSAVER;
    else if (g_strcmp0 (key, SCREENSAVER_LOCK_TIMEOUT) == 0)
        action_type = FD_IDLE_ACTION_SCREENSAVER_LOCK;
    else if (g_strcmp0 (key, SUSPEND_AC_TIMEOUT) == 0)
        action_type = FD_IDLE_ACTION_SUSPEND_AC;
    else if (g_strcmp0 (key, SUSPEND_BATTERY_TIMEOUT) == 0)
        action_type = FD_IDLE_ACTION_SUSPEND_BATTERY;
    else if (g_strcmp0 (key, DISPLAY_AC_TIMEOUT) == 0)
        action_type = FD_IDLE_ACTION_DISPLAY_AC;
    else if (g_strcmp0 (key, DISPLAY_BATTERY_TIMEOUT) == 0)
        action_type = FD_IDLE_ACTION_DISPLAY_BATTERY;

    return action_type;
}

static const gchar *
convert_action_type_to_key (FdIdleActionType action_type)
{
    const gchar *key = NULL;
    switch (action_type)
    {
        case FD_IDLE_ACTION_SCREENSAVER:
            key = SCREENSAVER_TIMEOUT;
            break;
        case FD_IDLE_ACTION_SCREENSAVER_LOCK:
            key = SCREENSAVER_LOCK_TIMEOUT;
            break;
        case FD_IDLE_ACTION_SUSPEND_AC:
            key = SUSPEND_AC_TIMEOUT;
            break;
        case FD_IDLE_ACTION_SUSPEND_BATTERY:
            key = SUSPEND_BATTERY_TIMEOUT;
            break;
        case FD_IDLE_ACTION_DISPLAY_AC:
            key = DISPLAY_AC_TIMEOUT;
            break;
        case FD_IDLE_ACTION_DISPLAY_BATTERY:
            key = DISPLAY_BATTERY_TIMEOUT;
            break;
        default:
            break;
    }
    return key;
}

static GSettings *
get_gsettings_from_action_type (FdIdleActionSettings *settings, 
                                FdIdleActionType action_type)
{
    FdIdleActionSettingsPrivate *priv = settings->priv;
    GSettings *gsettings = priv->power_gsettings;
    switch (action_type)
    {
        //case FD_IDLE_ACTION_SCREENSAVER:
        case FD_IDLE_ACTION_SCREENSAVER_LOCK:
            gsettings = priv->session_gsettings;
            break;
        default:
            break;
    }
    return gsettings;
}

static void
gsettings_on_changed  (GSettings *gsettings,
                       char      *key,
                       gpointer   user_data)
{
    FdIdleActionSettings *settings = FD_IDLE_ACTION_SETTINGS (user_data);
    FdIdleActionSettingsPrivate *priv = settings->priv;
    gint timeout = 0;
    FdIdleActionType action_type;
    action_type = convert_key_to_action_type (key);
    if (action_type == FD_IDLE_ACTION_NONE)
        return ;

    timeout = g_settings_get_uint (gsettings, key);
    gchar *schema_id;
    g_object_get (gsettings, "schema-id", &schema_id, NULL);
    g_debug ("Fd idle action settings(%s) changed, new timeout: %d", key, timeout);
    g_signal_emit (settings, 
                   settings_signals[CHANGED], 
                   0, 
                   action_type);
}

static void
fd_idle_action_settings_dispose (GObject *gobject)
{
    FdIdleActionSettings *settings = FD_IDLE_ACTION_SETTINGS (gobject);
    FdIdleActionSettingsPrivate *priv = settings->priv;

    if (priv->power_gsettings)
    {
        g_object_unref (priv->power_gsettings);
        priv->power_gsettings = NULL;
    }
    if (priv->session_gsettings)
    {
        g_object_unref (priv->session_gsettings);
        priv->session_gsettings = NULL;
    }
    G_OBJECT_CLASS (fd_idle_action_settings_parent_class)->dispose (gobject);
}

static void
fd_idle_action_settings_finalize (GObject *gobject)
{
    FdIdleActionSettings *settings = FD_IDLE_ACTION_SETTINGS (gobject);
    FdIdleActionSettingsPrivate *priv = settings->priv;

    G_OBJECT_CLASS (fd_idle_action_settings_parent_class)->finalize (gobject);
}

static void
fd_idle_action_settings_class_init (FdIdleActionSettingsClass *klass)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS (klass);

    gobject_class->finalize = fd_idle_action_settings_finalize;
    gobject_class->dispose = fd_idle_action_settings_dispose;

    settings_signals[CHANGED] = g_signal_new ("changed",
                                            G_TYPE_FROM_CLASS (klass),
                                            G_SIGNAL_RUN_LAST,
                                            NULL,
                                            NULL,
                                            NULL,
                                            g_cclosure_marshal_VOID__UINT,
                                            G_TYPE_NONE, 
                                            1, 
                                            G_TYPE_UINT);
}

static void
fd_idle_action_settings_init (FdIdleActionSettings *settings)
{
    FdIdleActionSettingsPrivate *priv;
    priv = settings->priv = fd_idle_action_settings_get_instance_private (settings);
    priv->power_gsettings = g_settings_new (POWER_SCHEMA_ID);
    g_signal_connect (priv->power_gsettings, 
                      "changed", 
                      G_CALLBACK (gsettings_on_changed), 
                      settings);
    priv->session_gsettings = g_settings_new (SCREENSAVER_SCHEMA_ID);
    g_signal_connect (priv->session_gsettings, 
                      "changed", 
                      G_CALLBACK (gsettings_on_changed), 
                      settings);
}

FdIdleActionSettings *
fd_idle_action_settings_new (void)
{
    FdIdleActionSettings *settings;
    settings = (FdIdleActionSettings *)g_object_new (FD_TYPE_IDLE_ACTION_SETTINGS, NULL);
    return settings;
}

guint
fd_idle_action_settings_get_timeout (FdIdleActionSettings *settings, 
                                     FdIdleActionType action)
{
    g_return_val_if_fail (FD_IS_IDLE_ACTION_SETTINGS (settings), 0);
    FdIdleActionSettingsPrivate *priv = settings->priv;
    const gchar *key;
    GSettings *gsettings;
    gsettings = get_gsettings_from_action_type (settings, action);
    key = convert_action_type_to_key (action);
    if (key)
        return g_settings_get_uint (gsettings, key);
    return 0;
}

gboolean
fd_idle_action_settings_set_timeout (FdIdleActionSettings *settings, 
                                     FdIdleActionType action, 
                                     guint timeout)
{
    g_return_val_if_fail (FD_IS_IDLE_ACTION_SETTINGS (settings), FALSE);
    FdIdleActionSettingsPrivate *priv = settings->priv;
    const gchar *key;
    gboolean ret = FALSE;
    GSettings *gsettings;
    gsettings = get_gsettings_from_action_type (settings, action);
    key = convert_action_type_to_key (action);
    if (key)
    {
        ret = g_settings_set_uint (gsettings, key, timeout);
        g_settings_sync ();
    }
    return ret;
}