#include <stdio.h>
#include "fd-login.h"

#define SYSTEMD_LOGIN1_BUS_NAME             "org.freedesktop.login1"
#define SYSTEMD_LOGIN1_MANAGER_OBJECT       "/org/freedesktop/login1"
#define SYSTEMD_LOGIN1_MANAGER_INTERFACE    "org.freedesktop.login1.Manager"

struct _FdLoginPrivate
{
    GDBusProxy  *login1_proxy;
};

enum
{
    CHANGED,
    PREPARE_FOR_SHUTDOWN,
    PREPARE_FOR_SLEEP,
    LAST_SIGNAL
};

static guint my_signals[LAST_SIGNAL] = { 0,};

G_DEFINE_TYPE_WITH_PRIVATE (FdLogin, fd_login, G_TYPE_OBJECT);

static gboolean
login_can_suspend_or_hibernate (FdLogin *self, const gchar *method_name)
{
    FdLoginPrivate *priv = self->priv;
    g_autoptr(GVariant) variant = NULL;
    g_autoptr(GError) error = NULL;
    const char *s;
    gboolean can = FALSE;
    if (!priv->login1_proxy)
        return can;

    variant = g_dbus_proxy_call_sync (priv->login1_proxy,
                                      method_name,
                                      NULL,
                                      G_DBUS_CALL_FLAGS_NONE,
                                      -1,
                                      NULL,
                                      &error);
    if (error)
    {
        g_warning ("Called login %s() failed: %s", method_name, error->message);
        return can;
    }
    g_variant_get (variant, "(&s)", &s);
    can = g_strcmp0 (s, "yes") == 0;
    return can;
}

static gboolean
check_acpi_supports (FdLogin *self, const gchar *action)
{
    gboolean has = TRUE;
    const char *file = "/etc/acpi/supports";
    char *check_content;
    FILE *stream = NULL;
    char buff[1024] = {0};
    const gchar *dbus_method;
    if (!g_file_test (file, G_FILE_TEST_EXISTS) || !action)
    {
        g_warning ("ACPI supports file is not found.");
        //ignore the file,if it doesn't exist.
        goto out;
    }
    check_content = g_strconcat ("head ", file, NULL);
    stream = popen (check_content, "r");
    g_free (check_content);
    if (stream == NULL)
    {
        goto out;
    }
    fread (buff, 1, sizeof (buff), stream);
    pclose (stream);
    if (buff != NULL)
    {
        if (strlen (buff) == 0 || strlen (buff) < strlen (action))
        {
            goto out;
        }
        if (strstr (buff, action) == NULL)
        {
            has = FALSE;
        }
    }

out:
    if (g_strcmp0 (action, "S3") == 0)
    {
        dbus_method = "CanSuspend";
    } 
    else if (g_strcmp0 (action, "S4") == 0)
    {
        dbus_method = "CanHibernate";
    }
    gboolean check_dbus = login_can_suspend_or_hibernate (self, dbus_method);
    g_message ("Check %s action: systemd_%s = %d, acpi_support = %d.", action, dbus_method, check_dbus, has);
    return has && check_dbus;
}

static void
login_hibernate_or_suspend_cb (GObject *source_object, 
                                 GAsyncResult *res, 
                                 gpointer user_data)
{
    g_autoptr (GVariant) result = NULL;
    g_autoptr(GError) error = NULL;

    result = g_dbus_proxy_call_finish (G_DBUS_PROXY (source_object), res, &error);
    if (error)
    {
        g_warning ("Called login hibernate or suspend failed: %s", error->message);
    }
}

static void
login1_proxy_on_signal_cb (GDBusProxy *proxy,
                            char       *sender_name,
                            char       *signal_name,
                            GVariant   *parameters,
                            gpointer    user_data)
{
    FdLogin *self = FD_LOGIN (user_data);
    FdLoginPrivate *priv = self->priv;
    gboolean p;
    g_debug ("Recieved signal from login1 proxy: %s", signal_name);
    if (g_strcmp0 (signal_name, "PrepareForShutdown") == 0)
    {
        g_variant_get (parameters, "(b)", &p);
        g_signal_emit (self, 
                       my_signals[PREPARE_FOR_SHUTDOWN], 
                       0, 
                       p);
    } 
    else if (g_strcmp0 (signal_name, "PrepareForSleep") == 0)
    {
        g_variant_get (parameters, "(b)", &p);
        g_signal_emit (self, 
                       my_signals[PREPARE_FOR_SLEEP], 
                       0, 
                       p);
    }
}

static void
fd_login_dispose (GObject *gobject)
{
    FdLogin *self = FD_LOGIN (gobject);
    FdLoginPrivate *priv = self->priv;

    if (priv->login1_proxy)
    {
        g_object_unref (priv->login1_proxy);
        priv->login1_proxy = NULL;
    }

    G_OBJECT_CLASS (fd_login_parent_class)->dispose (gobject);
}

static void
fd_login_finalize (GObject *gobject)
{
    FdLogin *self = FD_LOGIN (gobject);
    FdLoginPrivate *priv = self->priv;

    G_OBJECT_CLASS (fd_login_parent_class)->finalize (gobject);
}

static void
fd_login_class_init (FdLoginClass *klass)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS (klass);

    gobject_class->finalize = fd_login_finalize;
    gobject_class->dispose = fd_login_dispose;

    my_signals[CHANGED] = g_signal_new ("changed",
                                        G_TYPE_FROM_CLASS (klass),
                                        G_SIGNAL_RUN_LAST,
                                        NULL,
                                        NULL,
                                        NULL,
                                        g_cclosure_marshal_VOID__VOID,
                                        G_TYPE_NONE,
                                        0);
    my_signals[PREPARE_FOR_SHUTDOWN] = g_signal_new ("prepare-for-shutdown",
                                        G_TYPE_FROM_CLASS (klass),
                                        G_SIGNAL_RUN_LAST,
                                        NULL,
                                        NULL,
                                        NULL,
                                        g_cclosure_marshal_VOID__BOOLEAN,
                                        G_TYPE_NONE,
                                        1, 
                                        G_TYPE_BOOLEAN);
    my_signals[PREPARE_FOR_SLEEP] = g_signal_new ("prepare-for-sleep",
                                        G_TYPE_FROM_CLASS (klass),
                                        G_SIGNAL_RUN_LAST,
                                        NULL,
                                        NULL,
                                        NULL,
                                        g_cclosure_marshal_VOID__BOOLEAN,
                                        G_TYPE_NONE,
                                        1, 
                                        G_TYPE_BOOLEAN);
}

static void
fd_login_init (FdLogin *self)
{
    FdLoginPrivate *priv;
    GError *error = NULL;
    priv = self->priv = fd_login_get_instance_private (self);

    priv->login1_proxy = g_dbus_proxy_new_for_bus_sync (G_BUS_TYPE_SYSTEM,
                                    G_DBUS_PROXY_FLAGS_DO_NOT_AUTO_START,
                                    NULL,
                                    SYSTEMD_LOGIN1_BUS_NAME,
                                    SYSTEMD_LOGIN1_MANAGER_OBJECT,
                                    SYSTEMD_LOGIN1_MANAGER_INTERFACE,
                                    NULL,
                                    &error);
    if (error)
    {
        g_warning ("Connect systemd login1 failed: %s", error->message);
        g_error_free (error);
    }
    if (priv->login1_proxy)
    {
        g_signal_connect (priv->login1_proxy, 
                          "g-signal", 
                          G_CALLBACK (login1_proxy_on_signal_cb), 
                          self);
    }
}

FdLogin *
fd_login_get_default ()
{
    static FdLogin *login = NULL;
    if (login == NULL)
        login = (FdLogin *)g_object_new (FD_TYPE_LOGIN, NULL);
    return login;
}

void
fd_login_suspend (FdLogin *self)
{
    g_return_if_fail (FD_IS_LOGIN(self));
    FdLoginPrivate *priv = self->priv;
    if (!priv->login1_proxy)
        return;

    g_dbus_proxy_call (priv->login1_proxy,
                        "Suspend",
                        g_variant_new ("(b)", TRUE),
                        G_DBUS_CALL_FLAGS_NONE,
                        -1,
                        NULL,
                        (GAsyncReadyCallback) login_hibernate_or_suspend_cb,
                        self);
}

gboolean
fd_login_can_suspend (FdLogin *self)
{
    g_return_val_if_fail (FD_IS_LOGIN(self), FALSE);
    return check_acpi_supports (self, "S3");
}

void 
fd_login_hibernate (FdLogin *self)
{
    g_return_if_fail (FD_IS_LOGIN(self));
    FdLoginPrivate *priv = self->priv;
    if (!priv->login1_proxy)
        return;
    g_dbus_proxy_call (priv->login1_proxy,
                        "Hibernate",
                        g_variant_new ("(b)", TRUE),
                        G_DBUS_CALL_FLAGS_NONE,
                        -1,
                        NULL,
                        (GAsyncReadyCallback) login_hibernate_or_suspend_cb,
                        self);
}

gboolean 
fd_login_can_hibernate (FdLogin *self)
{
    g_return_val_if_fail (FD_IS_LOGIN(self), FALSE);

    return check_acpi_supports (self, "S4");
}