/**
 * SECTION: fd-session
 * @short_description: 当前会话管理器相关行为
 * @include: fd-session.h
 *
 * FdSessionPresenceClass, FdSessionInhibitor 类
 * 以及相关的dbus来处理登入登出会话进程.
 */
#include "fd-session.h"
#include "stdlib.h"
#include "stdio.h"

#define SESSION_BUS_NAME            "org.gnome.SessionManager"
#define SESSION_MANAGER_OBJECT      "/org/gnome/SessionManager"
#define SESSION_MANAGER_INTERFACE   "org.gnome.SessionManager"

#define SESSION_INHIBITOR_INTERFACE "org.gnome.SessionManager.Inhibitor"
#define SESSION_PRESENCE_OBJECT     "/org/gnome/SessionManager/Presence"
#define SESSION_PRESENCE_INTERFACE  "org.gnome.SessionManager.Presence"

/* Presence Object */
enum
{
    STATUS_CHANGED,
    SESSION_OVER,
    SESSION_RUNNING,
    ACTIVE_CHANGED,
    INHIBITED_ACTIONS_CHANGED,
    LAST_SIGNAL
};
static guint presence_signals[LAST_SIGNAL] = {0};

typedef struct _FdSessionPresence
{
    GObject parent_instance;

    GDBusProxy *proxy;
    guint32 status;
}FdSessionPresence;

G_DEFINE_TYPE (FdSessionPresence, fd_session_presence, G_TYPE_OBJECT);

static void
presence_on_status_changed (GDBusProxy *proxy,
                            gchar *sender,
                            gchar *signal_name,
                            GVariant *parameters,
                            FdSessionPresence *presence)
{
    g_debug ("Recieved signal from presence proxy: %s", signal_name);
    if (g_strcmp0 (signal_name, "StatusChanged") == 0)
    {
        GVariantIter iter;
        GVariant *child;
        guint status;

        g_variant_iter_init (&iter, parameters);
        child = g_variant_iter_next_value (&iter);
        status = g_variant_get_uint32 (child);
        g_variant_unref (child);
        presence->status = status;
        g_signal_emit (presence, 
                       presence_signals[STATUS_CHANGED], 
                       0, 
                       status);
    }
}

static void
fd_session_presence_finalize (GObject *gobject)
{
    G_OBJECT_CLASS (fd_session_presence_parent_class)->finalize (gobject);
}

static void
fd_session_presence_dispose (GObject *gobject)
{
    FdSessionPresence *presence = FD_SESSION_PRESENCE (gobject);

    g_clear_object (&presence->proxy);

    G_OBJECT_CLASS (fd_session_presence_parent_class)->dispose (gobject);
}

static void
fd_session_presence_class_init (FdSessionPresenceClass *klass)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS (klass);

    gobject_class->finalize = fd_session_presence_finalize;
    gobject_class->dispose = fd_session_presence_dispose;

    presence_signals[STATUS_CHANGED] = g_signal_new ("status-changed",
                                                     FD_TYPE_SESSION_PRESENCE,
                                                     G_SIGNAL_RUN_LAST,
                                                     0,
                                                     NULL,
                                                     NULL,
                                                     g_cclosure_marshal_VOID__UINT,
                                                     G_TYPE_NONE,
                                                     1,
                                                     G_TYPE_UINT);
}

static void
fd_session_presence_init (FdSessionPresence *presence)
{
    GError *error = NULL;
    GVariant *property;
    guint32 status = 0;
    //建立连接到 /org/gnome/SessionManager/Presence的代理
    presence->proxy = g_dbus_proxy_new_for_bus_sync (G_BUS_TYPE_SESSION,
                                                 G_DBUS_PROXY_FLAGS_DO_NOT_LOAD_PROPERTIES,
                                                 NULL,
                                                 SESSION_BUS_NAME,
                                                 SESSION_PRESENCE_OBJECT,
                                                 SESSION_PRESENCE_INTERFACE,
                                                 NULL,
                                                 &error);

    if (error)
    {
        g_warning ("Couldn't get org.gnome.SessionManager.Presence proxy: %s", error->message);
        g_error_free (error);
        return;
    }
    g_signal_connect (presence->proxy, 
                      "g-signal", 
                      G_CALLBACK (presence_on_status_changed), 
                      presence);

    property = g_dbus_proxy_get_cached_property (presence->proxy, "status");
    if (property)
        g_variant_get (property, "(u)", &status);
    presence->status = status;
}

/**
 * fd_session_presence_get_default:
 *
 * 新建一个 #FdSessionPresence 类的实例。
 *
 * Returns: 返回一个 #FdSessionPresence 的实例。
 */
FdSessionPresence *
fd_session_presence_get_default (void)
{
    static FdSessionPresence *presence = NULL;
    if (!presence)
    {
        presence = g_object_new (FD_TYPE_SESSION_PRESENCE, NULL);
    }
    return presence;
}

/**
 * fd_session_presence_get_status:
 * @presence: A #FdSessionPresence
 *
 * 这里通过dbus同步Proxy直接获取类返回值
 *
 * Returns (guint32): status
 */
guint32
fd_session_presence_get_status (FdSessionPresence *presence)
{
    g_return_val_if_fail (FD_IS_SESSION_PRESENCE (presence), 0);
    return presence->status;
}


/**
 * fd_session_presence_set_status:
 * @presence: A #FdSessionPresence
 * @status: 状态
 *
 * 这里通过dbus同步Proxy设置获取类返回值
 *
 * Returns: 设置成功返回 %TRUE, 否则返回 %FALSE
 */
gboolean
fd_session_presence_set_status (FdSessionPresence *presence, guint32 status)
{
    g_return_val_if_fail (FD_IS_SESSION_PRESENCE (presence), FALSE);
    g_autoptr(GError) error = NULL;
    g_autoptr (GVariant) retval;

    retval = g_dbus_proxy_call_sync (presence->proxy, 
                                     "SetStatus", 
                                     g_variant_new ("(u)", status), 
                                     G_DBUS_CALL_FLAGS_NONE, 
                                     -1, NULL, 
                                     &error);
    if (error)
    {
        g_warning ("Called %s method failed: %s",
                   g_dbus_proxy_get_interface_name (presence->proxy),
                   error->message);
        return FALSE;
    }

    g_dbus_proxy_set_cached_property (presence->proxy, "status", retval);
    return TRUE;
}

/* Inhibitor Object */
typedef struct _FdSessionInhibitor
{
    GObject parent_instance;
    GDBusProxy *proxy;

} FdSessionInhibitor;

G_DEFINE_TYPE (FdSessionInhibitor, fd_session_inhibitor, G_TYPE_OBJECT);

static void
fd_session_inhibitor_dispose (GObject *gobject)
{
    FdSessionInhibitor *inhibitor = FD_SESSION_INHIBITOR (gobject);

    if (inhibitor->proxy)
    {
        g_object_unref (inhibitor->proxy);
        inhibitor->proxy = NULL;
    }

    G_OBJECT_CLASS (fd_session_inhibitor_parent_class)->dispose (gobject);
}

static void
fd_session_inhibitor_class_init (FdSessionInhibitorClass *klass)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS(klass);
    //gobject_class->finalize = fd_session_inhibitor_finalize;
    gobject_class->dispose = fd_session_inhibitor_dispose;
}

static void
fd_session_inhibitor_init (FdSessionInhibitor *inhibitor)
{
    /* pass */
}

static gboolean
fd_session_inhibitor_setup (FdSessionInhibitor *inhibitor, gchar *object_path)
{
   g_autoptr(GError) error = NULL;
    inhibitor->proxy = g_dbus_proxy_new_for_bus_sync (G_BUS_TYPE_SESSION,
                                                G_DBUS_PROXY_FLAGS_DO_NOT_LOAD_PROPERTIES,
                                                NULL,
                                                SESSION_BUS_NAME,
                                                object_path,
                                                SESSION_INHIBITOR_INTERFACE,
                                                NULL,
                                                &error);
    if (error)
    {
        g_warning ("Create inhibitor dbus proxy failed: %s", error->message);
        return FALSE;
    }
    return TRUE;
}

FdSessionInhibitor *
fd_session_inhibitor_new (gchar *object_path)
{
    g_return_val_if_fail (g_variant_is_object_path (object_path), NULL);
    FdSessionInhibitor *inhibitor = g_object_new (FD_TYPE_SESSION_INHIBITOR, NULL);
    gboolean ret = fd_session_inhibitor_setup (inhibitor, object_path);
    if (!ret)
    {
        g_object_unref (inhibitor);
        return NULL;
    }
    return inhibitor;
}

gchar *
fd_session_inhibitor_get_app_id (FdSessionInhibitor *inhibitor)
{
    g_autoptr(GError) error = NULL;
    g_autoptr (GVariant) v = NULL;
    gchar *app_id = NULL, *new_app_id = NULL;

    v = g_dbus_proxy_call_sync (inhibitor->proxy, 
                                "GetAppId", 
                                NULL, 
                                G_DBUS_CALL_FLAGS_NONE, 
                                -1, 
                                NULL, 
                                &error);
    if (error)
    {
        g_warning ("Session inhibitor get app id failed:%s", error->message);
        return new_app_id;
    }

    g_variant_get (v, "(s)", &app_id);

    if (app_id && g_strcmp0 (app_id, ""))
        new_app_id = g_strdup (app_id);

    g_free (app_id);
    return new_app_id;
}

gchar *
fd_session_inhibitor_get_reason (FdSessionInhibitor *inhibitor)
{
    g_autoptr(GError) error = NULL;
    g_autoptr (GVariant) v;
    gchar *reason = NULL;

    v = g_dbus_proxy_call_sync (inhibitor->proxy, 
                                "GetReason", 
                                NULL, 
                                G_DBUS_CALL_FLAGS_NONE, 
                                -1, 
                                NULL, 
                                &error);
    if (error)
    {
        g_warning ("Session inhibitor get reason failed:%s", error->message);
        return reason;
    }

    g_variant_get (v, "(s)", &reason);

    return reason;
}

/* Session Object */
typedef struct _FdSession
{
    GObject parent;
    GDBusProxy *proxy;
    gboolean is_active;
    guint32 inhibited_actions;
} FdSession;

static guint session_signals[LAST_SIGNAL] = {0};

G_DEFINE_TYPE (FdSession, fd_session, G_TYPE_OBJECT)


static void
session_proxy_on_signal_cb (GDBusProxy *proxy,
                            char       *sender_name,
                            char       *signal_name,
                            GVariant   *parameters,
                            gpointer    user_data)
{
    FdSession *session = FD_SESSION (user_data);
    g_debug ("Recieved signal from session proxy: %s", signal_name);
    if (g_strcmp0 (signal_name, "SessionOver") == 0)
    {
        g_signal_emit (session, 
                       session_signals[SESSION_OVER], 
                       0);
    } 
    else if (g_strcmp0 (signal_name, "SessionRunning") == 0)
    {
        g_signal_emit (session, 
                       session_signals[SESSION_RUNNING], 
                       0);
    }
}

static void
session_proxy_on_properties_changed_cb (GDBusProxy *proxy,
                                        GVariant   *changed_properties,
                                        GStrv       invalidated_properties,
                                        gpointer    user_data)
{
    FdSession *session = FD_SESSION (user_data);
    GVariantIter iter;
    GVariant *value;
    gchar *key;

    g_variant_iter_init (&iter, changed_properties);
    while (g_variant_iter_loop (&iter, "{sv}", &key, &value))
    {
        if (g_strcmp0 (key, "SessionIsActive") == 0)
        {
            gboolean is_active = g_variant_get_boolean (value);
            g_message ("Session properties changed: %s = %d", key, is_active);
            session->is_active = is_active;
            g_signal_emit (session, 
                           session_signals[ACTIVE_CHANGED], 
                           0, 
                           is_active);
        } 
        else if (g_strcmp0 (key, "InhibitedActions") == 0)
        {
            guint32 inhibited_actions = g_variant_get_uint32 (value);
            g_message ("Session properties changed: %s = %u", key, inhibited_actions);
            session->inhibited_actions = inhibited_actions;
            g_signal_emit (session, 
                           session_signals[INHIBITED_ACTIONS_CHANGED], 
                           0, 
                           inhibited_actions);
        }
    }
}

static void
fd_session_dispose (GObject *gobject)
{
    FdSession *session = FD_SESSION (gobject);

    if (session->proxy)
    {
        g_object_unref (session->proxy);
        session->proxy = NULL;
    }
    G_OBJECT_CLASS (fd_session_parent_class)->dispose (gobject);
}

static void
fd_session_init (FdSession *session)
{
    GError *error = NULL;
    GVariant *v = NULL;
    session->is_active = FALSE;
    session->inhibited_actions = 0;

    session->proxy = g_dbus_proxy_new_for_bus_sync (G_BUS_TYPE_SESSION,
                              G_DBUS_PROXY_FLAGS_DO_NOT_AUTO_START,
                              NULL,
                              SESSION_BUS_NAME,
                              SESSION_MANAGER_OBJECT,
                              SESSION_MANAGER_INTERFACE,
                              NULL,
                              &error);
    if (error)
    {
        g_warning ("Connect session manager failed: %s", error->message);
        g_error_free (error);
    }
    if (session->proxy)
    {
        g_signal_connect (session->proxy, 
                          "g-signal", 
                          G_CALLBACK (session_proxy_on_signal_cb), 
                          session);
        g_signal_connect (session->proxy, 
                          "g-properties-changed", 
                          G_CALLBACK (session_proxy_on_properties_changed_cb), 
                          session);
        //init active status.
        v = g_dbus_proxy_get_cached_property (session->proxy,
                                              "SessionIsActive");
        if (v)
        {
            session->is_active = g_variant_get_boolean (v);
            g_variant_unref (v);
        }
        //init inhibited actions.
        v = g_dbus_proxy_get_cached_property (session->proxy,
                                              "InhibitedActions");
        if (v)
        {
            session->inhibited_actions = g_variant_get_uint32 (v);
            g_variant_unref (v);
        }
    }
}

static void
fd_session_class_init (FdSessionClass *klass)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS(klass);
    gobject_class->dispose = fd_session_dispose;

    session_signals[SESSION_OVER] = g_signal_new ("session-over",
                                                    FD_TYPE_SESSION,
                                                     G_SIGNAL_RUN_LAST,
                                                     0,
                                                     NULL,
                                                     NULL,
                                                     g_cclosure_marshal_VOID__VOID,
                                                     G_TYPE_NONE,
                                                     0);
    session_signals[SESSION_RUNNING] = g_signal_new ("session-running",
                                                    FD_TYPE_SESSION,
                                                     G_SIGNAL_RUN_LAST,
                                                     0,
                                                     NULL,
                                                     NULL,
                                                     g_cclosure_marshal_VOID__VOID,
                                                     G_TYPE_NONE,
                                                     0);
    session_signals[ACTIVE_CHANGED] = g_signal_new ("active-changed",
                                                    FD_TYPE_SESSION,
                                                     G_SIGNAL_RUN_LAST,
                                                     0,
                                                     NULL,
                                                     NULL,
                                                     g_cclosure_marshal_VOID__UINT,
                                                     G_TYPE_NONE,
                                                     1,
                                                     G_TYPE_UINT);
    session_signals[INHIBITED_ACTIONS_CHANGED] = g_signal_new ("inhibited-actions-changed",
                                                    FD_TYPE_SESSION,
                                                     G_SIGNAL_RUN_LAST,
                                                     0,
                                                     NULL,
                                                     NULL,
                                                     g_cclosure_marshal_VOID__UINT,
                                                     G_TYPE_NONE,
                                                     1,
                                                     G_TYPE_UINT);
}

/**
 * fd_session_get_default:
 *
 * 获取一个 #FdSession 类的实例。如果没有就新建一个。
 *
 * Return value: (transfer none): #FdSession
 */
FdSession *
fd_session_get_default (void)
{
    static FdSession *session = NULL;
    if (!session)
    {
        session = g_object_new (FD_TYPE_SESSION, NULL);
    }
    return session;
}

static void
session_logout_cb (GObject *source_object, 
                   GAsyncResult *res, 
                   gpointer user_data)
{
    g_autoptr (GVariant) result = NULL;
    g_autoptr(GError) error = NULL;

    result = g_dbus_proxy_call_finish (G_DBUS_PROXY (source_object), res, &error);
    if (error)
    {
        g_warning ("Called logout from session failed: %s", error->message);
    }
}

static void
session_shutdown_or_reboot_cb (GObject *source_object, 
                               GAsyncResult *res, 
                               gpointer user_data)
{
    g_autoptr (GVariant) result = NULL;
    g_autoptr(GError) error = NULL;

    result = g_dbus_proxy_call_finish (G_DBUS_PROXY (source_object), res, &error);
    if (error)
    {
        g_warning ("Called session shutdown or reboot failed: %s", error->message);
    }
}

gboolean
fd_session_is_active (FdSession *session)
{
    g_return_val_if_fail (FD_IS_SESSION(session), FALSE);
    return session->is_active;
}

gboolean 
fd_session_is_running (FdSession *session)
{
    g_return_val_if_fail (FD_IS_SESSION(session), FALSE);
    g_autoptr (GVariant) v = NULL;
    g_autoptr(GError) error = NULL;
    gboolean is_running = FALSE;
    if (!session->proxy)
        return is_running;

    v = g_dbus_proxy_call_sync (session->proxy,
                                "IsSessionRunning",
                                NULL,
                                G_DBUS_CALL_FLAGS_NONE,
                                -1, 
                                NULL,
                                &error);
    if (error)
    {
        g_warning ("Called session dbus method failed: %s", error->message);
        return is_running;
    }
    g_variant_get (v, "(b)", &is_running);
    return is_running;
}

const gchar *
fd_session_get_session_name (FdSession *session)
{
    g_return_val_if_fail (FD_IS_SESSION(session), NULL);
    g_autoptr (GVariant) v = NULL;
    const gchar *session_name = NULL;

    if (!session->proxy)
        return session_name;

    v = g_dbus_proxy_get_cached_property (session->proxy,
                                          "SessionName");
    if (v)
        session_name = g_variant_get_string (v, NULL);
    return session_name;
}

/**
 * fd_session_logout:
 * @mode: 模式
 *
 * 通过DBus代理获取当前模式进行注销;
 *
 */
void
fd_session_logout (FdSession *session, guint mode)
{
    g_return_if_fail (FD_IS_SESSION(session));
    if (!session->proxy)
        return;

    g_dbus_proxy_call (session->proxy,
                       "Logout",
                       g_variant_new ("(u)", mode),
                       G_DBUS_CALL_FLAGS_NONE,
                       -1, NULL,
                       session_logout_cb,
                       NULL);
}

/**
 * fd_session_shutdown:
 *
 * 通过session会话进行关机
 *
 */
void
fd_session_shutdown (FdSession *session)
{
    g_return_if_fail (FD_IS_SESSION(session));
    if (!session->proxy)
        return;
    g_dbus_proxy_call (session->proxy,
                       "Shutdown",
                       NULL,
                       G_DBUS_CALL_FLAGS_NONE,
                       -1,
                       NULL,
                       session_shutdown_or_reboot_cb,
                       NULL);
}

/**
 * fd_session_can_shutdown:
 *
 * 通过DBus获取当前代理的 "Canshutdown", 查询是否可以关机
 *
 * Return value: %TRUE 可以关机; %FALSE 则不可以.
 */
gboolean
fd_session_can_shutdown (FdSession *session)
{
    g_return_val_if_fail (FD_IS_SESSION(session), TRUE);
    if (!session->proxy)
        return TRUE;

    g_autoptr(GError) error = NULL;
    g_autoptr (GVariant) result = NULL;
    gboolean can_shutdown = TRUE;

    result = g_dbus_proxy_call_sync (session->proxy,
                                     "CanShutdown",
                                     NULL,
                                     G_DBUS_CALL_FLAGS_NONE,
                                     -1,
                                     NULL,
                                     &error);
    if (error)
    {
        g_warning ("Called session dbus method failed: %s", error->message);
        return can_shutdown;
    }
    g_variant_get (result, "(b)", &can_shutdown);
    return can_shutdown;
}

/**
 * fd_session_reboot:
 *
 * 通过session会话进行关机
 */
void
fd_session_reboot (FdSession *session)
{
    g_return_if_fail (FD_IS_SESSION(session));
    if (!session->proxy)
        return;
    g_dbus_proxy_call (session->proxy,
                       "Reboot",
                       NULL,
                       G_DBUS_CALL_FLAGS_NONE,
                       -1,
                       NULL,
                       session_shutdown_or_reboot_cb,
                       NULL);
}

/**
 * fd_session_get_inhibitors:
 *
 * 获取当前阻止注销、锁屏、关机等操作的应用
 *
 * Return value: (transfer full): 包含Inhibitor Dbus path 的数组
 */
char **
fd_session_get_inhibitors (FdSession *session)
{
    g_return_val_if_fail (FD_IS_SESSION(session), NULL);
    if (!session->proxy)
        return NULL;

    g_autoptr(GError) error = NULL;
    g_autoptr (GVariant) v = NULL;
    gchar **results= NULL;

    v = g_dbus_proxy_call_sync (session->proxy,
                                "GetInhibitors",
                                NULL,
                                G_DBUS_CALL_FLAGS_NONE,
                                -1,
                                NULL,
                                &error);
    if (error)
    {
        g_warning ("Called dbus method 'GetInhibitors' failed: %s", error->message);
        return NULL;
    }
    g_variant_get (v, "(^ao)", &results);
    if (results && g_strv_length (results) == 0)
    {
        g_strfreev (results);
        results = NULL;
    }
    return results;
}

guint32
fd_session_get_inhibited_actions (FdSession *session)
{
    g_return_val_if_fail (FD_IS_SESSION(session), 0);
    return session->inhibited_actions;
}
